"""Top-level package for airflint."""
from importlib.metadata import version
from os import getcwd
from os.path import dirname, join, realpath

__app_name__ = "airflint"
__version__ = version(__name__)
__location__ = realpath(join(getcwd(), dirname(__file__)))



from refactor import run

from airflint.rules.use_function_level_imports import UseFunctionLevelImports
from airflint.rules.use_jinja_variable_get import UseJinjaVariableGet


def main():
    run(
        rules=[
            UseFunctionLevelImports,
            UseJinjaVariableGet,
        ],
    )


if __name__ == "__main__":
    main()



import ast
from typing import Optional

from refactor import Representative
from refactor.context import ScopeInfo


class ImportFinder(Representative):
    """Find import by alias name."""

    def collect(self, name: str, scope: ScopeInfo) -> Optional[ast.ImportFrom]:
        return next(
            (
                node
                for node in ast.walk(self.context.tree)
                if isinstance(node, ast.ImportFrom)
                and any(alias.name == name for alias in node.names)
                and scope.can_reach(self.context["scope"].resolve(node))
            ),
            None,
        )



import ast
import importlib

from refactor import ReplacementAction, Rule
from refactor.context import Ancestry, Scope

from airflint.representatives.import_finder import ImportFinder


class UseJinjaVariableGet(Rule):
    """Replace `Variable.get("foo")` Calls through the jinja equivalent `{{ var.value.foo }}` if the variable is listed in `template_fields`."""

    context_providers = (Scope, Ancestry, ImportFinder)

    def match(self, node):
        assert (
            isinstance(node, ast.Call)
            and isinstance(node.func, ast.Attribute)
            and isinstance(node.func.value, ast.Name)
            and node.func.value.id == "Variable"
            and node.func.attr == "get"
            and isinstance(node.func.ctx, ast.Load)
            and isinstance(variable := node.args[0], ast.Constant)
        )
        assert (
            (parents := self.context["ancestry"].get_parents(node))
            and isinstance(operator_keyword := next(parents), ast.keyword)
            and isinstance(operator_call := next(parents), ast.Call)
            and isinstance(operator_call.func, ast.Name)
            and (
                import_node := self.context["import_finder"].collect(
                    operator_call.func.id,
                    scope=self.context["scope"].resolve(operator_call.func),
                )
            )
            and import_node.module
        )
        try:
            _module = importlib.import_module(import_node.module)
        except ImportError:
            return
        with open(_module.__file__) as file:
            module = ast.parse(file.read())
        assert any(
            isinstance(stmt, ast.AnnAssign)
            and isinstance(stmt.target, ast.Name)
            and stmt.target.id == "template_fields"
            and isinstance(stmt.value, ast.Tuple)
            and any(
                isinstance(elt, ast.Constant) and elt.value == operator_keyword.arg
                for elt in stmt.value.elts
            )
            for module_stmt in module.body
            if isinstance(module_stmt, ast.ClassDef)
            for stmt in module_stmt.body
        )

        return ReplacementAction(
            node,
            target=ast.Constant(value=f"{{{{ var.value.{variable.value} }}}}"),
        )



# Credits go to GitHub user @isidentical who provided most of the solution.
import ast

from refactor import Action, Rule, context
from refactor.context import ScopeType

from airflint.actions.move_statements import MoveStatementsAction


class UseFunctionLevelImports(Rule):
    """Replace top-level import by function-level import."""

    context_providers = (context.Scope,)

    def match(self, node: ast.AST) -> Action:
        # Instead of going import -> function, we are going
        # function -> import since this way it is easier to
        # do the refactorings one by one and finally remove
        # all unused imports at the end.
        assert isinstance(node, ast.FunctionDef)

        inlined_imports = []
        for identifier in ast.walk(node):
            # Find all used identifiers inside this function.
            if not (
                isinstance(identifier, ast.Name)
                and isinstance(identifier.ctx, ast.Load)
            ):
                continue

            # And try to resolve their scope. Each scope has its own parent
            # (unless it is a module), so we are simply going back until we
            # find the definition for the selected identifier.
            scope = self.context["scope"].resolve(identifier)
            while not (definitions := scope.definitions.get(identifier.id, [])):
                scope = scope.parent
                if scope is None:
                    # There might be some magic, so let's not
                    # forget the chance of running out of scopes.
                    break

            # If any of the definitions (there might be multiple of them)
            # we found matches an import, we'll filter them out.
            imports = [
                definition
                for definition in definitions
                if isinstance(definition, ast.Import)
            ]
            # And we'll ensure this import is originating from the global scope.
            if imports and scope.scope_type is ScopeType.GLOBAL:
                inlined_imports.append(imports[0])

        # We want this rule to only run if there is at least 1 inlined import.
        assert len(inlined_imports) >= 1

        # We'll select the first statement, which will act like an anchor to us
        # when we are inserting.
        first_stmt = node.body[0]
        return MoveStatementsAction(first_stmt, inlined_imports)



# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['airflint', 'airflint.actions', 'airflint.representatives', 'airflint.rules']

package_data = \
{'': ['*']}

install_requires = \
['refactor>=0.4.3,<0.5.0']

entry_points = \
{'console_scripts': ['airflint = airflint.__main__:main']}

setup_kwargs = {
    'name': 'airflint',
    'version': '0.2.0a0',
    'description': 'Enforce Best Practices for all your Airflow DAGs. ⭐',
    'long_description': '# airflint\n\n[![pre-commit.ci status](https://results.pre-commit.ci/badge/github/feluelle/airflint/main.svg)](https://results.pre-commit.ci/latest/github/feluelle/airflint/main)\n![test workflow](https://github.com/feluelle/airflint/actions/workflows/test.yml/badge.svg)\n![codeql-analysis workflow](https://github.com/feluelle/airflint/actions/workflows/codeql-analysis.yml/badge.svg)\n[![codecov](https://codecov.io/gh/feluelle/airflint/branch/main/graph/badge.svg?token=J8UEP8IVY4)](https://codecov.io/gh/feluelle/airflint)\n[![PyPI version](https://img.shields.io/pypi/v/airflint)](https://pypi.org/project/airflint/)\n[![License](https://img.shields.io/pypi/l/airflint)](https://github.com/feluelle/airflint/blob/main/LICENSE)\n[![PyPI - Python Version](https://img.shields.io/pypi/pyversions/airflint)](https://pypi.org/project/airflint/)\n[![PyPI version](https://img.shields.io/pypi/dm/airflint)](https://pypi.org/project/airflint/)\n\n> Enforce Best Practices for all your Airflow DAGs. ⭐\n\n## 🧑\u200d🏫 Rules\n\n- [x] Use function-level imports instead of top-level imports[^1] (see [Top level Python Code](https://airflow.apache.org/docs/apache-airflow/stable/best-practices.html#top-level-python-code))\n- [x] Use jinja macro instead of `Variable.get` (see [Airflow Variables](https://airflow.apache.org/docs/apache-airflow/stable/best-practices.html#airflow-variables))\n\n[^1]: There is a PEP for [Lazy Imports](https://peps.python.org/pep-0690/) targeted to arrive in Python 3.12 which would supersede this rule.\n\n_based on official [Best Practices](https://airflow.apache.org/docs/apache-airflow/stable/best-practices.html)_\n\n## 🚀 Get started\n\nTo install it from [PyPI](https://pypi.org/) run:\n\n```console\npip install airflint\n```\n\nThen just call it like this:\n\n![usage](assets/images/usage.png)\n\n### pre-commit\n\nAlternatively you can add the following repo to your `pre-commit-config.yaml`:\n\n```yaml\n  - repo: https://github.com/feluelle/airflint\n    rev: v0.1.2-alpha\n    hooks:\n      - id: airflint\n        args: ["-a"]  # Use -a for replacing inplace\n```\n',
    'author': 'Felix Uellendall',
    'author_email': 'feluelle@pm.me',
    'maintainer': None,
    'maintainer_email': None,
    'url': None,
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'entry_points': entry_points,
    'python_requires': '>=3.9,<4.0',
}


setup(**setup_kwargs)



# Credits go to GitHub user @isidentical who provided most of the solution.
import ast
from dataclasses import dataclass

from refactor import Action, Context, common
from refactor.ast import split_lines


@dataclass
class MoveStatementsAction(Action):
    """Move statements from their old location to the node's location."""

    statements: list[ast.stmt]

    def apply(self, context: Context, source: str) -> str:
        # We need a custom action that defines a custom apply()
        # method since this use case is not natively supported.

        # What apply() does is very basic, it splits the whole
        # source into smaller pieces (lines) and inserts the
        # newly built source code into the appropriate place.
        lines = split_lines(source)

        # First we need to find where the initial anchor starts,
        # and we'll extract the indentation from there. So all
        # the new nodes will align with this indentation level.
        start_line = lines[self.node.lineno - 1]
        indentation, start_prefix = common.find_indent(
            start_line[: self.node.col_offset],
        )

        # Then we'll build the actual source code from scratch.
        new_source = ""
        for statement in self.statements:
            new_source += context.unparse(statement) + "\n"

        # And split it into lines as well.
        replacement = split_lines(new_source)

        # For each line in this code, we'll apply *THE SAME* indentation level
        # as the anchor node.
        replacement.apply_indentation(indentation, start_prefix=start_prefix)

        # And insert these lines just on top of the actual anchor node.
        anchor = self.node.lineno - 1
        for line in reversed(replacement):
            lines.insert(anchor, line)

        # We remove the old statements.
        for line in split_lines(new_source):
            lines.remove(line)

        # Finally we'll merge everything together!
        return lines.join()
